from typing import Any, Callable, overload

from ._types import Comparable, Numeric, T


@overload
def wrapped_summation() -> Callable[..., Numeric]:
    ...


@overload
def wrapped_summation(start: T) -> Callable[..., T]:
    ...


def wrapped_summation(start=0):
    """
    Wrapped summation function.
    """

    def inner_summation(*values: T) -> T:
        return sum(values, start=start)

    return inner_summation


@overload
def wrapped_multiplication() -> Callable[..., Numeric]:
    ...


@overload
def wrapped_multiplication(start: T) -> Callable[..., T]:
    ...


def wrapped_multiplication(start=1):
    """
    Wrapped multiplication function.
    """

    def inner_multiplication(*values: T) -> T:
        result = start
        for number in values:
            result *= number
        return result

    return inner_multiplication


def wrapped_chained_comparison(
    comparison_func: Callable[[Comparable, Comparable], bool]
):
    """
    Wrapped chained comparison.
    """

    def chained_comparison(*values: Comparable) -> bool:
        """
        Return `values[0] < values[1] < ... < values[-1]`.
        """
        current = values[0]
        for i in range(1, len(values)):
            if comparison_func(current, values[i]):
                current = values[i]
            else:
                return False
        else:
            return True

    return chained_comparison


def chained_keyword_and(*values: Any) -> Any:
    """
    Return `values[0] and values[1] and ... and values[-1]`.
    """
    current = values[0]
    for i in range(1, len(values)):
        if not current:
            break
        current = current and values[i]
    return current


def chained_keyword_or(*values: Any) -> Any:
    """
    Return `values[0] or values[1] or ... or values[-1]`.
    """
    current = values[0]
    for i in range(1, len(values)):
        if current:
            break
        current = current or values[i]
    return current


def wrapped_chained_general_operation(
    binary_func: Callable[[Any, Any], Any]
) -> Callable[..., Any]:
    """
    Wrapped chained general operation.
    """

    def chained_general_operation(*values: Any) -> Any:
        current = values[0]
        for i in range(1, len(values)):
            current = binary_func(current, values[i])
        return current

    return chained_general_operation



from numbers import Number
from typing import Any, Callable, Protocol, SupportsFloat, TypeVar, Union

Numeric = Union[Number, SupportsFloat]

T = TypeVar("T")
S = TypeVar("S")
R = TypeVar("R")


class BoundMethod(Protocol[S, R]):
    """
    Represents bound method. Used for typing hints.
    """

    __func__: Callable[..., R]
    __self__: S

    def __call__(self, *args: Any, **kwargs: Any) -> R:
        ...


class Comparable(Protocol):
    """
    Represents comparable object, but without `__eq__` and `__ne__`.
    """

    def __lt__(self, other) -> bool:
        ...

    def __le__(self, other) -> bool:
        ...

    def __gt__(self, other) -> bool:
        ...

    def __ge__(self, other) -> bool:
        ...


class ComparisonFunc(Protocol):
    """
    Represents comparison function.
    """

    def __call__(self, arg1: Any, arg2: Any) -> bool:
        ...



import asyncio
from typing import Any, AsyncGenerator, Generator

from .graph import toposort
from .indicators import AbstractIndicator


def generate_sync(
    **indicators: AbstractIndicator,
) -> Generator[dict[str, Any], None, None]:
    """
    Synchronously generate all indicators by flow.
    """
    extended_toposorted: list[list[AbstractIndicator]] = toposort(*indicators.values())
    try:
        while True:
            for batched_indicators in extended_toposorted:
                for indicator in batched_indicators:
                    indicator.update_single()
            yield {
                indicator_name: indicator(0)
                for indicator_name, indicator in indicators.items()
            }
    except StopIteration:
        pass


async def generate_async(
    **indicators: AbstractIndicator,
) -> AsyncGenerator[dict[str, Any], None]:
    """
    Asynchronously generate all indicators by flow.
    """
    extended_toposorted: list[list[AbstractIndicator]] = toposort(*indicators.values())
    try:
        while True:
            for batched_indicators in extended_toposorted:
                done, pending = await asyncio.wait(
                    [
                        asyncio.create_task(indicator.update_single_async())
                        for indicator in batched_indicators
                    ],
                    return_when=asyncio.FIRST_EXCEPTION,
                )
                while done:
                    task = done.pop()
                    exc = task.exception()
                    if exc:
                        for running_task in pending:
                            running_task.cancel()
                        raise exc
            yield {
                indicator_name: indicator(0)
                for indicator_name, indicator in indicators.items()
            }
    except StopAsyncIteration:
        pass



from sortedcontainers import SortedList

from .._types import Numeric, T
from .base import AbstractHistoryTrackingIndicator, AbstractIndicator


class SimpleMovingAverage(AbstractHistoryTrackingIndicator[Numeric]):
    """
    SMA indicator. Exact formula of this indicator is
    `SMA(source, len) = sum(source, len) / (len - none_count)`
    where `None` is treated as zero and
    `none_count` is number of `None`s in tracking history.
    """

    def __init__(self, indicator: AbstractIndicator, length: int, **kwargs) -> None:
        super().__init__(indicator, tracking_lengths=(length,), **kwargs)
        self._none_count = sum(
            1
            for i in range(self._tracking_lengths[0])
            if self.pre_requisites[0](i) is None
        )
        self._sum = 0

    def remove_tail(self) -> None:
        value_removal = self.pre_requisites[0](self._tracking_lengths[0])
        self._sum -= value_removal or 0  # type: ignore
        if value_removal is None:
            self._none_count -= 1

    def add_head(self) -> None:
        value_new = self.pre_requisites[0](0)
        self._sum += value_new or 0
        if value_new is None:
            self._none_count += 1

    def update_single(self) -> None:
        super().update_single()
        self.set_value(
            self._sum / (self._tracking_lengths[0] - self._none_count)
            if self._tracking_lengths[0] > self._none_count
            else self._default_value
        )


class SimpleMovingVariance(AbstractHistoryTrackingIndicator[Numeric]):
    """
    Variance indicator. Exact formula of this indicator is
    `Var(source, len) = variance(filter(None, source), len - none_count)`.
    If all values in tracking history are `None`, then produces `None`.
    """

    def __init__(self, indicator: AbstractIndicator, length: int, **kwargs) -> None:
        super().__init__(indicator, tracking_lengths=(length,), **kwargs)
        self._none_count = sum(
            1
            for i in range(self._tracking_lengths[0])
            if self.pre_requisites[0](i) is None
        )
        self._sum = 0
        self._square_sum = 0

    def remove_tail(self) -> None:
        value_removal = self.pre_requisites[0](self._tracking_lengths[0])
        self._sum -= value_removal or 0
        self._square_sum -= (value_removal or 0) ** 2
        if value_removal is None:
            self._none_count -= 1

    def add_head(self) -> None:
        value_new = self.pre_requisites[0](0)
        self._sum += value_new or 0
        self._square_sum += (value_new or 0) ** 2
        if value_new is None:
            self._none_count += 1

    def update_single(self) -> None:
        super().update_single()
        length = self._tracking_lengths[0] - self._none_count
        self.set_value(
            self._square_sum / length - (self._sum / length) ** 2
            if length > 0
            else self._default_value
        )


class SimpleHistoricalStats(AbstractHistoryTrackingIndicator[dict[str, Numeric]]):
    """
    Historical statistical indicator. Generates historical
    minimum, maximum, and median of only dependency.
    The reason why statistic is saved in child class
    while history is saved in parent class is, otherwise
    statistic is not reliable when the older value exists.
    """

    def __init__(self, indicator: AbstractIndicator, length: int, **kwargs) -> None:
        super().__init__(indicator, tracking_lengths=(length,), **kwargs)
        self._indicator_statistics: SortedList[T] = SortedList()

    def remove_tail(self) -> None:
        value_removal = self.pre_requisites[0](self._tracking_lengths[0])
        if value_removal is not None:
            self._indicator_statistics.remove(value_removal)

    def add_head(self) -> None:
        value_new = self.pre_requisites[0](0)
        if value_new is not None:
            self._indicator_statistics.add(value_new)

    def _get_median(self) -> Numeric:
        length = len(self._indicator_statistics)
        if length % 2 == 1:
            return self._indicator_statistics[(length - 1) // 2]
        else:
            return (
                self._indicator_statistics[length // 2]
                + self._indicator_statistics[length // 2 - 1]
            ) / 2

    def update_single(self) -> None:
        super().update_single()
        self.set_value(
            {
                "min": self._indicator_statistics[0],
                "max": self._indicator_statistics[-1],
                "median": self._get_median(),
            }
            if self._indicator_statistics
            else self._default_value
        )



class IndicatorManagementError(Exception):
    """
    Abstract class of all errors in `indicator_management`.
    """


class CannotResolveIndicatorGraph(IndicatorManagementError):
    """
    Raised when the indicator dependencies
    graph cannot be resolved.
    """



from __future__ import annotations

from collections import deque
from functools import wraps
from numbers import Number
from typing import (
    Any,
    AsyncGenerator,
    AsyncIterable,
    Callable,
    Generic,
    Iterable,
    Iterator,
    Optional,
    Protocol,
    TypeVar,
    Union,
    cast,
)

from .._types import BoundMethod, ComparisonFunc, Numeric, T
from ..utils import (
    chained_keyword_and,
    chained_keyword_or,
    wrapped_chained_comparison,
    wrapped_chained_general_operation,
    wrapped_multiplication,
    wrapped_summation,
)

IOPR_co = TypeVar("IOPR_co", covariant=True)


class IndicatorOperationProtocol(Protocol[IOPR_co]):
    def __call__(
        self, *indicator_or_numbers: Union[AbstractIndicator, Numeric], **kwargs
    ) -> IOPR_co:
        raise NotImplementedError


def indicatorized_arguments(func: Callable[..., T]) -> IndicatorOperationProtocol[T]:
    """
    Convert a function which accepts indicators or numbers.
    In resulting function, all number arguments
    will be changed into `ConstantIndicator`.
    """

    @wraps(func)
    def inner_function(
        *indicator_or_numbers: Union[AbstractIndicator, Numeric], **kwargs
    ) -> T:
        result = []
        for obj in indicator_or_numbers:
            if isinstance(obj, AbstractIndicator):
                result.append(obj)
            elif isinstance(obj, Number):
                result.append(ConstantIndicator(obj))
            else:
                raise TypeError
        return func(*result, **kwargs)

    return inner_function


class AbstractIndicator(Generic[T]):
    """
    Abstract class of all indicators.
    """

    global_nonce: int = 0

    def __init__(
        self,
        *pre_requisites: AbstractIndicator,
        default_value: Optional[T] = None,
        history_length: int = 1,
    ) -> None:
        self.id = AbstractIndicator.global_nonce
        AbstractIndicator.global_nonce += 1

        self._default_value = default_value
        self.indicator_history: deque[Optional[T]] = deque(
            (None for _ in range(history_length - 1)),
            maxlen=history_length,
        )

        self._post_dependencies: list[AbstractIndicator] = []

        self.pre_requisites: tuple[AbstractIndicator, ...] = pre_requisites
        assert len(self.pre_requisites) == len(set(self.pre_requisites))
        for pre_requisite in self.pre_requisites:
            pre_requisite._post_dependencies.append(self)
        self.set_value(self._default_value)

    # =================================================================================
    # Indicator / History related

    def set_value(self, v: Optional[T]):
        """
        Set the new value of this indicator.
        """
        self.indicator_history.append(v)

    def resize_history(self, history_length: int, increase_only: bool = True):
        """
        Resize indicator history by given `history_length`.
        If `increase_only` is True, then the resulting length will only be increased.
        """
        if not increase_only or history_length > len(self.indicator_history):
            self.indicator_history = deque(
                self.indicator_history,
                maxlen=history_length,
            )
            self.indicator_history.extendleft(
                (None for _ in range(history_length - len(self.indicator_history)))
            )

    def __call__(self, index: int) -> Optional[T]:
        return self.indicator_history[-1 - index]

    # =================================================================================
    # Arithmetic operations

    def __add__(self, other) -> OperationIndicator:
        return summation(self, other, start=0)

    def __radd__(self, other) -> OperationIndicator:
        return summation(other, self, start=0)

    def __sub__(self, other) -> OperationIndicator:
        return subtraction(self, other)

    def __rsub__(self, other) -> OperationIndicator:
        return subtraction(other, self)

    def __mul__(self, other) -> OperationIndicator:
        return multiplication(self, other, start=1)

    def __rmul__(self, other) -> OperationIndicator:
        return multiplication(other, self, start=1)

    def __truediv__(self, other) -> OperationIndicator:
        return division(self, other)

    def __rtruediv__(self, other) -> OperationIndicator:
        return division(other, self)

    def __pos__(self) -> AbstractIndicator:
        return self

    def __neg__(self) -> OperationIndicator:
        return self * -1

    def __pow__(self, other) -> OperationIndicator:
        return power(self, other)

    def __rpow__(self, other) -> OperationIndicator:
        return power(other, self)

    # =================================================================================
    # Comparison operations

    def __lt__(self, other) -> OperationIndicator[bool]:
        return less(self, other)

    def __le__(self, other) -> OperationIndicator[bool]:
        return less_or_equal(self, other)

    def __gt__(self, other) -> OperationIndicator[bool]:
        return greater(self, other)

    def __ge__(self, other) -> OperationIndicator[bool]:
        return greater_or_equal(self, other)

    # =================================================================================
    # Logical operations

    def __and__(self, other) -> OperationIndicator:
        return and_operator(self, other, safe_none=False)

    def __or__(self, other) -> OperationIndicator:
        return or_operator(self, other, safe_none=False)

    def __xor__(self, other) -> OperationIndicator:
        return xor_operator(self, other, safe_none=False)

    # =================================================================================
    # Other operations which produces indicators

    def __getitem__(self, index: Any) -> OperationIndicator:
        return index_access(self, index=index)

    # =================================================================================
    # Extra magic methods

    def __hash__(self) -> int:
        return hash(self.id)

    def __repr__(self):
        return "<Indicator id %d :: Type %s :: History %s :: %d dependencies(%s)>" % (
            self.id,
            str(type(self)).split(".")[-1],
            self.indicator_history,
            len(self.pre_requisites),
            ", ".join(str(indicator.id) for indicator in self.pre_requisites),
        )

    # =================================================================================
    # Helper APIs

    def is_basis(self) -> bool:
        """
        Check if this indicator does not require any pre-requisites.
        """
        return not bool(self.pre_requisites)

    def get_post_dependencies(self) -> set[AbstractIndicator]:
        """
        Get set of post dependencies.
        """
        return set(self._post_dependencies)

    # =================================================================================
    # Update APIs

    async def update_single_async(self) -> None:
        """
        Asynchronously update current indicator.
        """
        self.update_single()

    def update_single(self) -> None:
        """
        Calculate and update new indicator value based on given `sources`.
        """
        raise NotImplementedError


AI = TypeVar("AI", bound=AbstractIndicator)


def default_if_none_in_pre_requisites(
    method: BoundMethod[AI, None]
) -> Callable[[], None]:
    """
    Decorate this to `Indicator.update_single` if
    you want to force `self.set_value(self._default_value)`
    when any pre-requisite indicator have `None` value.
    """

    def inner_method() -> None:
        self = method.__self__
        if any((indicator(0) is None) for indicator in self.pre_requisites):
            self.set_value(self._default_value)
        else:
            method()

    return inner_method


class RawSeriesIndicator(AbstractIndicator[T]):
    """
    Represents an indicator which is updated by
    synchronous raw value stream.
    """

    def __init__(self, *, raw_values: Iterable[Optional[T]], **kwargs) -> None:
        super().__init__(default_value=None, **kwargs)
        self._raw_values: Iterator[Optional[T]] = iter(raw_values)

    async def update_single_async(self) -> None:
        try:
            self.update_single()
        except StopIteration as exc:
            raise StopAsyncIteration(*exc.args).with_traceback(exc.__traceback__)

    def update_single(self) -> None:
        self.set_value(next(self._raw_values))


class AsyncRawSeriesIndicator(AbstractIndicator[T]):
    """
    Represents an indicator which is updated by
    asynchronous raw value stream.
    """

    def __init__(
        self,
        *,
        raw_values: AsyncIterable[Optional[T]],
        **kwargs,
    ) -> None:
        super().__init__(default_value=None, **kwargs)
        self._raw_values: AsyncGenerator[Optional[T], None] = (
            value async for value in raw_values
        )

    async def update_single_async(self) -> None:
        self.set_value(await self._raw_values.__anext__())


class ConstantIndicator(AbstractIndicator[T]):
    """
    Represents an constant indicator.
    """

    def __init__(self, constant: T) -> None:
        super().__init__(default_value=constant, history_length=0)

    def update_single(self) -> None:
        pass

    def __call__(self, index: int) -> Optional[T]:
        return self._default_value


N_or_AIN = Union[Numeric, AbstractIndicator[Numeric]]


class OperationIndicator(AbstractIndicator[T]):
    """
    Base of all operational indicators.
    """

    def __init__(
        self,
        *pre_requisites: AbstractIndicator,
        operation: Callable[..., Optional[T]],
        safe_none: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(*pre_requisites, **kwargs)
        self._operation = operation
        if safe_none:
            diniprs = default_if_none_in_pre_requisites
            self.update_single = diniprs(self.update_single)  # type: ignore

    def update_single(self) -> None:
        self.set_value(
            self._operation(
                *(pre_requisite(0) for pre_requisite in self.pre_requisites)
            )
        )


@indicatorized_arguments
def summation(
    *indicators: AbstractIndicator, start: T, **kwargs
) -> OperationIndicator[T]:
    """
    Summation of indicators.
    """
    return OperationIndicator(
        *indicators,
        operation=wrapped_summation(start),
        **kwargs,
    )


@indicatorized_arguments
def multiplication(
    *indicators: AbstractIndicator, start: T, **kwargs
) -> OperationIndicator[T]:
    """
    Multiplication of indicators.
    """
    return OperationIndicator(
        *indicators, operation=wrapped_multiplication(start), **kwargs
    )


@indicatorized_arguments
def subtraction(
    indicator1: AbstractIndicator, indicator2: AbstractIndicator, **kwargs
) -> OperationIndicator:
    """
    Subtraction of two indicators.
    """
    return OperationIndicator(
        indicator1, indicator2, operation=(lambda x, y: x - y), **kwargs
    )


@indicatorized_arguments
def division(
    indicator1: AbstractIndicator, indicator2: AbstractIndicator, **kwargs
) -> OperationIndicator:
    """
    Division of two indicators.
    """
    return OperationIndicator(
        indicator1, indicator2, operation=(lambda x, y: x / y if y else None), **kwargs
    )


@indicatorized_arguments
def power(
    indicator1: AbstractIndicator, indicator2: AbstractIndicator, **kwargs
) -> OperationIndicator:
    """
    Power operation of two indicators.
    """
    return OperationIndicator(
        indicator1, indicator2, operation=(lambda x, y: x**y), **kwargs
    )


@indicatorized_arguments
def less(*indicators: AbstractIndicator, **kwargs) -> OperationIndicator[bool]:
    """
    Generates `indicators[0] < indicators[1] < ... < indicators[-1]`.
    """
    return OperationIndicator(
        *indicators,
        operation=wrapped_chained_comparison(cast(ComparisonFunc, lambda x, y: x < y)),
        **kwargs,
    )


@indicatorized_arguments
def less_or_equal(*indicators: AbstractIndicator, **kwargs) -> OperationIndicator[bool]:
    """
    Generates `indicators[0] <= indicators[1] <= ... <= indicators[-1]`.
    """
    return OperationIndicator(
        *indicators,
        operation=wrapped_chained_comparison(cast(ComparisonFunc, lambda x, y: x <= y)),
        **kwargs,
    )


@indicatorized_arguments
def greater(*indicators: AbstractIndicator, **kwargs) -> OperationIndicator[bool]:
    """
    Generates `indicators[0] > indicators[1] > ... > indicators[-1]`.
    """
    return OperationIndicator(
        *indicators,
        operation=wrapped_chained_comparison(cast(ComparisonFunc, lambda x, y: x > y)),
        **kwargs,
    )


@indicatorized_arguments
def greater_or_equal(
    *indicators: AbstractIndicator, **kwargs
) -> OperationIndicator[bool]:
    """
    Generates `indicators[0] >= indicators[1] >= ... >= indicators[-1]`.
    """
    return OperationIndicator(
        *indicators,
        operation=wrapped_chained_comparison(cast(ComparisonFunc, lambda x, y: x >= y)),
        **kwargs,
    )


@indicatorized_arguments
def and_keyword(*indicators, **kwargs) -> OperationIndicator[bool]:
    """
    Generates `indicators[0] and indicators[1] and ... and indicators[-1]`.
    """
    return OperationIndicator(*indicators, operation=chained_keyword_and, **kwargs)


@indicatorized_arguments
def or_keyword(*indicators, **kwargs) -> OperationIndicator[bool]:
    """
    Generates `indicators[0] or indicators[1] or ... or indicators[-1]`.
    """
    return OperationIndicator(*indicators, operation=chained_keyword_or, **kwargs)


@indicatorized_arguments
def and_operator(*indicators, **kwargs) -> OperationIndicator:
    """
    Generates `indicators[0] & indicators[1] & ... & indicators[-1]`.
    """
    return OperationIndicator(
        *indicators,
        operation=wrapped_chained_general_operation(lambda x, y: x & y),
        **kwargs,
    )


@indicatorized_arguments
def or_operator(*indicators, **kwargs) -> OperationIndicator:
    """
    Generates `indicators[0] | indicators[1] | ... | indicators[-1]`.
    """
    return OperationIndicator(
        *indicators,
        operation=wrapped_chained_general_operation(lambda x, y: x | y),
        **kwargs,
    )


@indicatorized_arguments
def xor_operator(*indicators, **kwargs) -> OperationIndicator:
    """
    Generaters `indicators[0] ^ indicators[1] ^ ... ^ indicators[-1]`. (xor operator)
    """
    return OperationIndicator(
        *indicators,
        operation=wrapped_chained_general_operation(lambda x, y: x ^ y),
        **kwargs,
    )


def booleanize(indicator: AbstractIndicator, **kwargs) -> OperationIndicator:
    """
    Generates `bool(indicator)`.
    """
    return OperationIndicator(indicator, operation=(lambda x: bool(x)), **kwargs)


def index_access(
    indicator: AbstractIndicator, *, index: Any, **kwargs
) -> OperationIndicator:
    """
    Index access operation of an indicator.
    """
    return OperationIndicator(
        indicator, operation=(lambda x: x[index] if x is not None else None), **kwargs
    )


class AbstractHistoryTrackingIndicator(AbstractIndicator[T]):
    """
    Abstract base of all history-tracking indicators.
    """

    def __init__(
        self, *indicators: AbstractIndicator, tracking_lengths: Iterable[int], **kwargs
    ) -> None:
        super().__init__(*indicators, **kwargs)

        self._tracking_lengths: tuple[int, ...] = tuple(tracking_lengths)
        for indicator, length in zip(indicators, self._tracking_lengths, strict=True):
            if length <= 0:
                raise ValueError("Non-positive length %d given" % (length,))
            indicator.resize_history(length + 1, increase_only=True)

    def remove_tail(self) -> None:
        """
        Perform action related to removal of old values.
        """
        raise NotImplementedError

    def add_head(self) -> None:
        """
        Perform action related to addition of new values.
        """
        raise NotImplementedError

    def update_single(self) -> None:
        self.remove_tail()
        self.add_head()



from .base import AbstractIndicator, OperationIndicator, indicatorized_arguments


@indicatorized_arguments
def minimum(*indicators: AbstractIndicator, **kwargs) -> OperationIndicator:
    """
    Minimum indicator. Generates minimum of all dependencies'.
    """
    return OperationIndicator(*indicators, operation=min, **kwargs)


@indicatorized_arguments
def maximum(*indicators: AbstractIndicator, **kwargs) -> OperationIndicator:
    """
    Maximum indicator. Generates maximum of all dependencies'.
    """
    return OperationIndicator(*indicators, operation=max, **kwargs)



from .base import AbstractIndicator


class AgingIndicator(AbstractIndicator[int]):
    """
    Aging indicator. If dependency of this indicator's boolean conversion
    is True, then increase the age. Otherwise, set the age to zero.
    """

    def __init__(self, indicator: AbstractIndicator, **kwargs) -> None:
        super().__init__(indicator, default_value=0)

    def update_single(self) -> None:
        value = self.pre_requisites[0](0)
        current_age = self(0)
        self.set_value(current_age + 1 if value else 0)  # type: ignore



from .errors import CannotResolveIndicatorGraph
from .indicators import AbstractIndicator


def all_pre_requisites(*indicators: AbstractIndicator) -> set[AbstractIndicator]:
    """
    Return set of all pre-requisites (including ancestors) required by given
    indicators. This method does not detect cyclic dependencies.
    """
    visited_indicators: set[AbstractIndicator] = set()

    def dfs(indicator: AbstractIndicator):
        if indicator in visited_indicators:
            return
        visited_indicators.add(indicator)
        for pre_requisite in indicator.pre_requisites:
            dfs(pre_requisite)

    for indicator in indicators:
        dfs(indicator)
    return visited_indicators


def toposort(*indicators: AbstractIndicator) -> list[list[AbstractIndicator]]:
    """
    Return topological sorted list of given indicators and its ancestors.
    You can concurrently update all indicators in same list element in result.
    """
    expanded_indicators = all_pre_requisites(*indicators)
    result: list[list[AbstractIndicator]] = []
    removed: set[AbstractIndicator] = set()

    graph_forward = {
        indicator: {
            next_indicator
            for next_indicator in indicator.get_post_dependencies()
            if next_indicator in expanded_indicators
        }
        for indicator in expanded_indicators
    }
    graph_backward = {
        indicator: set(indicator.pre_requisites) for indicator in expanded_indicators
    }

    result.append([])
    for indicator in expanded_indicators:
        if not graph_backward[indicator]:
            result[0].append(indicator)
            removed.add(indicator)
    if not result[0]:
        raise CannotResolveIndicatorGraph(
            "There is no basis on given indicators. "
            "Perhaps you made an cyclic dependencies?"
        )

    while result[-1]:
        result.append([])
        for indicator in result[-2]:
            for next_indicator in graph_forward[indicator]:
                if next_indicator in removed:
                    continue
                graph_backward[next_indicator].remove(indicator)
                if not graph_backward[next_indicator]:
                    result[-1].append(next_indicator)
                    removed.add(next_indicator)

    if len(removed) != len(expanded_indicators):
        raise CannotResolveIndicatorGraph(
            "There are some remaining unvisited indicators. "
            "Perhaps you made an cyclic dependencies?"
        )

    result.pop()
    return result



from math import cos as _cos
from math import log as _log
from math import sin as _sin
from math import tan as _tan

from .._types import Numeric
from .base import AbstractIndicator, OperationIndicator, indicatorized_arguments


@indicatorized_arguments
def sin(indicator: AbstractIndicator[Numeric], **kwargs) -> OperationIndicator:
    """
    Sine of indicator.
    """
    return OperationIndicator(indicator, operation=_sin, **kwargs)


@indicatorized_arguments
def cos(indicator: AbstractIndicator[Numeric], **kwargs) -> OperationIndicator:
    """
    Cosine of indicator.
    """
    return OperationIndicator(indicator, operation=_cos, **kwargs)


@indicatorized_arguments
def tan(indicator: AbstractIndicator[Numeric], **kwargs) -> OperationIndicator:
    """
    Tangent of indicator.
    """
    return OperationIndicator(indicator, operation=_tan, **kwargs)


@indicatorized_arguments
def log(
    indicator_up: AbstractIndicator[Numeric],
    indicator_base: AbstractIndicator[Numeric],
    **kwargs
) -> OperationIndicator:
    """
    Logarithm of indicator.
    """
    return OperationIndicator(indicator_up, indicator_base, operation=_log, **kwargs)



from .._types import Numeric
from .base import AbstractIndicator
from .statistical import SimpleMovingAverage, SimpleMovingVariance


def bollinger_band(
    base_indicator: AbstractIndicator[Numeric],
    length: int,
    stdev_multiplier: float = 2.0,
) -> tuple[AbstractIndicator[Numeric], ...]:
    """
    Return upper bound, middle, and lower bound of Bollinger Band.
    """
    assert stdev_multiplier > 0
    moving_average = SimpleMovingAverage(base_indicator, length)
    bandwidth = SimpleMovingVariance(base_indicator, length) ** 0.5 * stdev_multiplier
    return moving_average + bandwidth, moving_average, moving_average - bandwidth


class ExponentialMovingAverage(AbstractIndicator[Numeric]):
    """
    EMA indicator. Exact formula of this indicator refers to
    [here](https://www.investopedia.com/terms/m/movingaverage.asp).
    """

    def __init__(
        self, indicator: AbstractIndicator, forward_weight: float = 2 / 21, **kwargs
    ) -> None:
        super().__init__(indicator, **kwargs)
        assert 0 < forward_weight < 1
        self._forward_weight = forward_weight

    def update_single(self) -> None:
        value = self.pre_requisites[0](0)
        prev_value = self(0)
        if value is None:
            self.set_value(self._default_value)
        elif prev_value is None:
            self.set_value(value)
        else:
            self.set_value(
                prev_value * (1 - self._forward_weight)  # type: ignore
                + value * self._forward_weight
            )



from .aging import AgingIndicator
from .base import (
    AbstractHistoryTrackingIndicator,
    AbstractIndicator,
    AsyncRawSeriesIndicator,
    ConstantIndicator,
    RawSeriesIndicator,
    and_keyword,
    and_operator,
    booleanize,
    division,
    greater,
    greater_or_equal,
    index_access,
    less,
    less_or_equal,
    multiplication,
    or_keyword,
    or_operator,
    power,
    subtraction,
    summation,
    xor_operator,
)
from .comparisons import maximum, minimum
from .financial import ExponentialMovingAverage, bollinger_band
from .mathematical import cos, log, sin, tan
from .statistical import (
    SimpleHistoricalStats,
    SimpleMovingAverage,
    SimpleMovingVariance,
)
